@@ -24,7 +24,7 @@ callr <- function(expr, envir = parent.frame(), substitute = TRUE,
   if (substitute) expr <- substitute(expr)
 
   if (is.null(workers)) workers <- availableCores()
-  stopifnot(length(workers) == 1L, is.numeric(workers),
+  stop_if_not(length(workers) == 1L, is.numeric(workers),
             is.finite(workers), workers >= 1L)
 
   future <- CallrFuture(expr = expr, envir = envir, substitute = FALSE,
